////////////////////////////////////////////////////////////////////////////
//
// COPYRIGHT ANSYS. ALL RIGHTS RESERVED.
//
// $LastChangedDate$
// $LastChangedRevision$
// $LastChangedBy$
//
// Author: ltomaso
//
// Decription : implementation of IOptimizationMethod and the CSharpOptimizer
//
////////////////////////////////////////////////////////////////////////////

using System;
using System.Collections.Generic;
using System.Linq;
using Ans.ComponentSystem;
using Ansys.DesignXplorer.API.Common;
using Ansys.DesignXplorer.API.Optimization;

namespace Ansys.DesignXplorer.CSharpOptimizerExample
{
    #region private enumerations

    // input parameter type
    enum MyVariableType
    {
        // defined by a range [lowerbound;upperbound] where lowerbound and upperbound are real values
        Continuous = 0,
        // defined by a sorted list L={v_1;v_2;...;v_N} where L[i] is a real value for i=1..N, and L[i]<L[i+1] for i=1..N-1
        Continuouslist = 1,
        // defined by a list L={v_1;v_2;...;v_N} where L[i] is an integer value for i=1..N, and L[i]!=L[j] for i,j=1..N and i!=j
        Integerlist = 2
    }

    // objective type
    enum MyObjectiveType
    {
        Minimize = 0,
        Maximize = 1,
        SeekTarget = 2
    }

    // constraint type
    enum MyConstraintType
    {
        LessThan = 0,
        GreaterThan = 1,
        EqualTo = 2,
        Between = 3
    }

    #endregion private enumerations

    // ScreningOptimizerCSharpExample generates a set of random points, and sort them with a cost function for example to extract the best candidate points
    //
    // Input Settings:
    // - Optimization Problem:
    //      - optimization domain (input variables)
    //      - objectives
    //      - constraints
    //      - output (output parameter can have objective or/and constraint) but it's not mandatory (for more details see AddOutput method)
    // - Optimization settings:
    //      - random generator seed 
    //      - number of sample points to generate
    //      - maximum number of candidate points
    //
    // Output Settings:
    // - number of evaluations
    // - candidate points
    public class CSharpOptimizer : IOptimizationMethod, IGridSupport
    {
        // the method's name should not countain empty space
        public const string MethodName = "CSharpOptimizer";

        #region private members

        // Optimization Services: provided by DesignXplorer
        // IOptimizationServices is used to:
        // - submit one or several optimization points to calculate
        // - promote optimization point as history point
        // - publish log message
        // - publish progress
        IOptimizationServices _optServices = null;

        // My Optimization Settings
        int _numberOfSamples = 0;
        int _maximumNumberOfCandidates = 0;
        int _numberOfEvaluations = 0;
        // Candidate points
        List<IOptimizationPoint> _candidates = new List<IOptimizationPoint>();

        // My Optimization problem
        // List of parameters id
        List<string> _parameters = new List<string>();
        // List of objectives id
        List<string> _objectives = new List<string>();
        // List of constraint id
        List<string> _constraints = new List<string>();

        // Dictionary of parameter types
        Dictionary<string, MyVariableType> _parametertype = new Dictionary<string, MyVariableType>();
        // Dictionary of objective types
        Dictionary<string, MyObjectiveType> _objectivetype = new Dictionary<string, MyObjectiveType>();
        // Dictionary of constraint types
        Dictionary<string, MyConstraintType> _constrainttype = new Dictionary<string, MyConstraintType>();

        // Dictionary of parameter values
        // _parametersvalues[id] has 2 values (lower and upper bounds) when _parametertype[id] == Continuous
        // _parametersvalues[id] has a sorted list of real values when _parametertype[id] == Continuouslist
        // _parametersvalues[id] has a list of integer values when _parametertype[id] == Integerlist
        Dictionary<string, List<object>> _parametersvalues = new Dictionary<string, List<object>>();

        // Dictionary of objective values
        // the list _objectivesvalues[id] is empty when _objectivetype[id] == Minimize or _objectivetype[id] == Maximize
        // the list _objectivesvalues[id] contains the target value when _objectivetype[id] == Minimize or _objectivetype[id] == SeekTarget
        Dictionary<string, List<object>> _objectivesvalues = new Dictionary<string, List<object>>();

        // Dictionary of constraint values
        // the list _constraintsvalues[id] contains the lower bound when _constrainttype[id] == LessThan
        // the list _constraintsvalues[id] contains the upper bound when _constrainttype[id] == GreaterThan
        // the list _constraintsvalues[id] contains the target value when _constrainttype[id] == EqualTo
        // the list _constraintsvalues[id] contains the lower bound and the upper bound when _constrainttype[id] == Between
        Dictionary<string, List<object>> _constraintsvalues = new Dictionary<string, List<object>>();

        // True when there is one parameter relationship enabled at least
        private bool _hasParameterRelationship = false;

        // random generator seed 
        int _seedValue = 0;
        // random generator
        Random _random = null;
        // random point index
        int _ptIdx = 0;

        #endregion private members

        #region Constructors

        public CSharpOptimizer()
        {
        }

        #endregion Constructors

        #region IOptimizationMethod

        // Services are initialized by DesignXplorer at the beginning of the Optimization run
        public IOptimizationServices Services
        {
            set
            {
                _optServices = value;
            }
        }

        public object get_Setting(string propertyName)
        {
            object propertyValue;
            switch (propertyName)
            {
                case "MyNumberOfSamples":
                    propertyValue = _numberOfSamples;
                    break;
                case "MyMaximumNumberOfCandidates":
                    propertyValue = _maximumNumberOfCandidates;
                    break;
                case "MyNumberOfEvaluations":
                    propertyValue = _numberOfEvaluations;
                    break;
                case "MySeedValue":
                    propertyValue = _seedValue;
                    break;
                default:
                    propertyValue = null;
                    break;
            }

            return propertyValue;
        }

        public void put_Setting(string propertyName, object propertyValue)
        {
            switch (propertyName)
            {
                case "MyNumberOfSamples":
                    _numberOfSamples = (int)propertyValue;
                    break;
                case "MyMaximumNumberOfCandidates":
                    _maximumNumberOfCandidates = (int)propertyValue;
                    break;
                case "MySeedValue":
                    _seedValue = (int)propertyValue;
                    _random = new Random(_seedValue);
                    break;
            }
        }

        // returns best candidate points
        public IListCpp Candidates
        {
            get
            {
                if (_candidates == null || _candidates.Count == 0)
                    return null;

                IListCpp points = new VariantColl();
                int count = _candidates.Count;
                for (int ic = 0; ic < count; ic++)
                    points.Add(_candidates[ic]);
                return points;
            }
        }

        // definition of double variable
        // certain optimizer starts from an initial point (gradient based method for example)
        // initialvalue corresponds to the current value of the variable for the initial point
        // Ignore initialvalue if the optimizer does not start from initial point
        public void AddDoubleVariable(string id, double lowerbound, double upperbound, double initialvalue)
        {
            _parameters.Add(id);
            _parametertype.Add(id, MyVariableType.Continuous);
            List<object> listofvalues = new List<object>();
            listofvalues.Add(lowerbound);
            listofvalues.Add(upperbound);
            _parametersvalues.Add(id, listofvalues);
        }

        // definition of integer list variable
        // certain optimizer starts from an initial point (gradient based method for example)
        // initialvalue corresponds to the current value of the variable for the initial point
        // Ignore initialvalue if the optimizer does not start from initial point
        public void AddIntegerListVariable(string id, IListCpp values, int initialvalue)
        {
            _parameters.Add(id);
            _parametertype.Add(id, MyVariableType.Integerlist);
            List<object> listofvalues = new List<object>();
            int count = values.Count;
            for (int i = 0; i < count; i++ )
                listofvalues.Add(values[i]);
            _parametersvalues.Add(id, listofvalues);
        }

        // definition of double list variable
        // certain optimizer starts from an initial point (gradient based method for example)
        // initialvalue corresponds to the current value of the variable for the initial point
        // Ignore initialvalue if the optimizer does not start from initial point
        public void AddDoubleListVariable(string id, IListCpp values, double initialvalue)
        {
            _parameters.Add(id);
            _parametertype.Add(id, MyVariableType.Continuouslist);
            List<object> listofvalues = new List<object>();
            int count = values.Count;
            for (int i = 0; i < count; i++)
                listofvalues.Add(values[i]);
            _parametersvalues.Add(id, listofvalues);
        }

        public void AddGridInterval(string id, double gridinterval)
        {
            if (!_parameters.Contains(id))
                return;

            MyVariableType type;
            if (!_parametertype.TryGetValue(id, out type))
                return;

            if (type != MyVariableType.Continuous)
                return;

            _parametersvalues[id].Add(gridinterval);
        }

        // definition of output parameter
        // the problem can have several output parameters, and only one objective
        // Example:
        //
        // you minimize the deformation of a beam and you have 2 others outputs:the mass and the stress of the beam
        // only output "deformation" is minimized but you need to know mass and stress values associated to this deformation
        // that's why 3 output parameters are defined : the deformation, the mass and the stress
        //
        // when any point is calculated, DesignXplorer set values on all output parameters in the optimization point
        // you can store all output values, 
        // or store only objectives and constraint values and ignore output parameters without objectives and constraints
        //
        // for your information when you publish history point, if an output value is missing in the optimization point
        // then the history chart associated to this parameter will be wrong
        // if you publish an optimization point calculated by DesignXplorer as history point, all output parameter values are defined in the optimization point
        // if you publish an optimization point calculated on your side (approximated with an internal response surface for example), 
        // you need to know all output parameter if you want to define all output parameter values in the optimization point 
        public void AddOutput(string outputid)
        {
            // nothing
        }

        public void AddCustomVariableProperty(string variableid, string propertyName, object propertyValue)
        {
            // nothing
        }

        public void AddParameterRelationship(string variableid, string leftexpresssion, string rightexpression, enumParameterRelationshipType type)
        {
            _hasParameterRelationship = true;
        }

        // definition of objective
        // goalvalue is only used when objective type is SeekTarget
        // in this example, importance is ignored
        public void AddObjective(string objectiveid, enumGoalType type, double goalvalue, double importance)
        {
            if (type == enumGoalType.eGT_NoPreference)
                return;

            _objectives.Add(objectiveid);
            List<object> values = new List<object>();
            switch (type)
            {
                case enumGoalType.eGT_MinimumPossible:
                    _objectivetype.Add(objectiveid, MyObjectiveType.Minimize);
                    _objectivesvalues.Add(objectiveid, values);
                    break;
                case enumGoalType.eGT_MaximumPossible:
                    _objectivetype.Add(objectiveid, MyObjectiveType.Maximize);
                    _objectivesvalues.Add(objectiveid, values);
                    break;
                case enumGoalType.eGT_SeekTarget:
                    _objectivetype.Add(objectiveid, MyObjectiveType.SeekTarget);
                    values.Add(goalvalue);
                    _objectivesvalues.Add(objectiveid, values);
                    break;
            }
        }

        // definition of constraint
        // constraintvalue1 corresponds to
        // - lower bound if constraint type is eCT_LessThanTarget
        // - upper bound if constraint type is eCT_GreaterThanTarget
        // - target value if constraint type is eCT_NearTarget
        // - lower bound if constraint type is eCT_InsideBounds
        // constraintvalue2 corresponds to
        // - upper bound if constraint type is eCT_InsideBounds
        //
        // importance and strictconstraint are ignored in this example.
        public void AddConstraint(string constraintid, enumConstraintType type, object constraintvalue1, object constraintvalue2, double importance, bool strictconstraint)
        {
            if (type == enumConstraintType.eCT_NoPreference)
                return;

            _constraints.Add(constraintid);
            List<object> values = new List<object>();
            switch (type)
            {
                case enumConstraintType.eCT_LessThanTarget:
                    _constrainttype.Add(constraintid, MyConstraintType.LessThan);
                    values.Add(constraintvalue1);
                    _constraintsvalues.Add(constraintid, values);
                    break;
                case enumConstraintType.eCT_GreaterThanTarget:
                    _constrainttype.Add(constraintid, MyConstraintType.GreaterThan);
                    values.Add(constraintvalue1);
                    _constraintsvalues.Add(constraintid, values);
                    break;
                case enumConstraintType.eCT_NearTarget:
                    _constrainttype.Add(constraintid, MyConstraintType.EqualTo);
                    values.Add(constraintvalue1);
                    _constraintsvalues.Add(constraintid, values);
                    break;
                case enumConstraintType.eCT_InsideBounds:
                    _constrainttype.Add(constraintid, MyConstraintType.Between);
                    values.Add(constraintvalue1);
                    values.Add(constraintvalue2);
                    _constraintsvalues.Add(constraintid, values);
                    break;
            }
        }

        public void AddCustomObjectiveProperty(string variableid, string propertyName, object propertyValue)
        {
            // nothing
        }

        public void AddCustomConstraintProperty(string variableid, string propertyName, object propertyValue)
        {
            // nothing
        }

        // CanRun is called just before to run the optimizer, to check the definition of the optimization problem
        // if the optimizer can not be run, you can specify the reason in the returned message
        // this message will be displayed in the message view of the interface
        public bool CanRun(out string message)
        {
            // check of the number of variables
            if (_parameters.Count == 0)
            {
                message = "You need to specify one parameter at least.";
                return false;
            }

            // check of the number of objectives
            if (_objectives.Count == 0)
            {
                message = "You need to specify one objective at least.";
                return false;
            }

            // check the consistency between the maximum number of candidates and the number of samples
            if (_maximumNumberOfCandidates <= 0 || _maximumNumberOfCandidates > _numberOfSamples)
            {
                message = "Maximum Number of Candidates must be positive and lower than or equal to the Number of Samples.";
                return false;
            }

            // check the number of sample points to generate
            if (_numberOfSamples <= 0 || _numberOfSamples > 10000)
            {
                message = "Number of Samples must be positive and lower than or equal to 10000.";
                return false;
            }

            message = string.Empty;
            return true;
        }

        // Optimization
        public void Run()
        {
            // initialization of the list of candidate points
            _candidates = null;

            // Publication of log message
            _optServices.PublishLogMessage("Generation of random points...\n");
            // generation of random points
            IListCpp points = new VariantColl();
            int iNumFeasiblePts = 0;
            for (int i = 0; i < _numberOfSamples * 100 && iNumFeasiblePts < _numberOfSamples; i++)
            {
                IOptimizationPoint point = GenerateRandomPoint();
                if (_hasParameterRelationship)
                {
                    _optServices.CalculateParameterRelationshipsFromPoint(point);
                    if (point.State == enumPointState.ePS_Invalid
                        || point.State == enumPointState.ePS_Failed)
                        continue;
                }
                points.Add(point);
                iNumFeasiblePts++;
            }

            if (iNumFeasiblePts == 0)
            {
                _optServices.PublishLogMessage("No feasible point!\n");
                return;
            }

            _optServices.PublishLogMessage("Calculation of points...\n");
            // split the points in up to 10 sets
            int nbsets = (10 > iNumFeasiblePts) ? iNumFeasiblePts : 10;
            int nbpointsbyset = iNumFeasiblePts / nbsets;
            IListCpp pointsToCalculate = new VariantColl();
            int ind = 0;
            int iset = 1;
            while (ind < iNumFeasiblePts)
            {
                pointsToCalculate.Clear();
                for (int j = 0; j < nbpointsbyset && ind < iNumFeasiblePts; ++j, ++ind)
                {
                    pointsToCalculate.Add(points[ind]);
                }

                // Calculation of points
                _optServices.CalculatePoints(pointsToCalculate);
                // Publish progress : the full optimization corresponds to 100 units
                // we consider (int this case), that 10% of the optimization has been achieved
                _optServices.PublishProgress(string.Format("Set #{0} calculated", iset), 10);
                iset++;

                // stop the processing, if users stops the optimization
                if (_optServices.Stopped)
                    break;
            }

            // sort of the best candidates
            Dictionary<string, double> costfunctionpts = new Dictionary<string, double>(points.Count);
            IListCpp sortedpoints = new VariantColl();
            int pointsCount = points.Count;
            for (int i = 0; i < pointsCount; i++)
            {
                IOptimizationPoint point = (IOptimizationPoint) points[i];
                // eOPS_UpToDate means the calculation of point is achieved without any error
                // all output values, objectives values and constraints values have been evaluated successfully
                if (point.State == enumPointState.ePS_UpToDate)
                {
                    // The point is displayed as History Point
                    _optServices.PushHistoryPoint(point);

                    double cost = CostFunction(point);
                    costfunctionpts.Add(point.ID, cost);

                    int ip = 0;
                    for (; ip < sortedpoints.Count; ip++)
                    {
                        IOptimizationPoint pt = (IOptimizationPoint) sortedpoints[ip];
                        if (cost < costfunctionpts[pt.ID])
                            break;
                    }

                    sortedpoints.Insert(ip, point);

                }
                // eOPS_OutOfDate means the calculation of point is not achieved (if users stops the optimization before the end)
                if (point.State != enumPointState.ePS_OutOfDate)
                    _numberOfEvaluations++;
            }

            _optServices.PublishLogMessage("Sort of points : generation of candidate points...\n");
            // Extract the best candidate points of the sorted sample points
            _candidates = new List<IOptimizationPoint>(_maximumNumberOfCandidates);
            for (int ip = 0; ip < _maximumNumberOfCandidates; ip++)
            {
                IOptimizationPoint pt = (IOptimizationPoint)sortedpoints[ip];
                _candidates.Add(pt);
                _optServices.PushHistoryPoint(pt);
            }
        }

        // this optimizer returns only candidate points as sample points for the post processing
        // these points are used to build chart results
        public enumPostProcessingType PostProcessingTypes
        {
            get { return enumPostProcessingType.ePPT_Candidates; }
        }

        public IListCpp Samples
        {
            get { throw new NotImplementedException(); }
        }

        #endregion IOptimizationMethod

        #region private methods

        // creates and generates a random point
        // only input variables values are defined
        private IOptimizationPoint GenerateRandomPoint()
        {
            IOptimizationPoint point = new DXOptimizationPoint();
            point.ID = "DP" + Convert.ToString(_ptIdx, System.Globalization.CultureInfo.InvariantCulture);
            _ptIdx++;

            if (_random == null)
                _random = new Random(_seedValue);

            int paramCount = _parameters.Count;
            for (int ip=0; ip<paramCount ; ip++)
            {
                string param = _parameters[ip];
                switch (_parametertype[param])
                {
                    case MyVariableType.Continuous:
                        {
                            double rand = _random.NextDouble();
                            double lb = (double) _parametersvalues[param][0];
                            double ub = (double) _parametersvalues[param][1];
                            double val = lb + rand * (ub - lb);
                            if (_parametersvalues[param].Count > 2) //= grid interval
                            {
                                double tol = (double)_parametersvalues[param][2];
                                val = RoundToNearestGridValue(val, lb, ub, tol);
                            }

                            point.put_Value(param, val);
                        }
                        break;
                    case MyVariableType.Continuouslist:
                        {
                            int ub = (int) _parametersvalues[param].Count();
                            int idx = _random.Next(ub);

                            double val = (double) _parametersvalues[param][idx];
                            point.put_Value(param, val);
                        }
                        break;
                    case MyVariableType.Integerlist:
                        {
                            int ub = (int) _parametersvalues[param].Count();
                            int idx = _random.Next(ub);

                            int val = (int) _parametersvalues[param][idx];
                            point.put_Value(param, val);
                        }
                        break;

                }
            }
            return point;
        }

        // Computes cost function of the current objective
        private double ComputeObjectiveFunction(IOptimizationPoint point, string objectiveid)
        {
            double objFunc = 0.0;
            if (!_objectives.Contains(objectiveid))
                return objFunc;

            switch (_objectivetype[objectiveid])
            {
                case MyObjectiveType.Minimize:
                    objFunc = (double)point.get_Value(objectiveid);
                    break;
                case MyObjectiveType.Maximize:
                    objFunc = -((double)point.get_Value(objectiveid));
                    break;
                case MyObjectiveType.SeekTarget:
                    objFunc = Math.Abs(((double)_objectivesvalues[objectiveid][0]) - ((double)point.get_Value(objectiveid)));
                    break;
            }

            return objFunc;
        }

        // Compute cost function of the current constraint
        private double ComputeConstraintFunction(IOptimizationPoint point, string constraintid)
        {
            double constrFunc = 0.0;
            if (!_constraints.Contains(constraintid))
                return constrFunc;

            switch (_constrainttype[constraintid])
            {
                case MyConstraintType.LessThan:
                    constrFunc = ((double)point.get_Value(constraintid)) - ((double)_constraintsvalues[constraintid][0]);
                    break;
                case MyConstraintType.GreaterThan:
                    constrFunc = ((double)_constraintsvalues[constraintid][0]) - ((double)point.get_Value(constraintid));
                    break;
                case MyConstraintType.EqualTo:
                    constrFunc = Math.Abs(((double)point.get_Value(constraintid)) - ((double)_constraintsvalues[constraintid][0]));
                    break;
                case MyConstraintType.Between:
                    {
                        double c1 = ((double)_constraintsvalues[constraintid][0]) - ((double)point.get_Value(constraintid));
                        double c2 = ((double)point.get_Value(constraintid)) - ((double)_constraintsvalues[constraintid][1]);
                        constrFunc = Math.Max(c1, c2);
                    }
                    break;
            }

            return constrFunc;
        }

        // the global cost function is the sum of cost function of objectives and constraints
        private double CostFunction(IOptimizationPoint point)
        {
            double cost1 = _objectives.Sum(id => ComputeObjectiveFunction(point, id));
            double cost2 = _constraints.Sum(id => ComputeConstraintFunction(point, id));
            double cost = cost1 + cost2;
            return cost;
        }

        private double RoundToNearestGridValue(double val, double lowerbound, double upperbound, double gridinterval)
        {
            if (gridinterval <= 0)
                return val;

            if (val < lowerbound)
                return lowerbound;

            if (val > upperbound)
                return upperbound;

            double dStep = (val - lowerbound) / gridinterval;
            int iStep = (int) Math.Round(dStep);

            double gridvalue = Math.Min(lowerbound + gridinterval * iStep, upperbound);
            return gridvalue;
        }

        #endregion private methods
    }
}
